@@ -22,19 +22,19 @@
 #include "apr_file_io.h"
 #include "apr_thread_mutex.h"
 #include "apr_poll.h"
+#include "apr_pools.h"
 
 #ifdef HAVE_OPENSSL
 #include "ssl_private.h"
 
 static jclass byteArrayClass;
+static jclass stringClass;
 
 static apr_status_t ssl_context_cleanup(void *data)
 {
     tcn_ssl_ctxt_t *c = (tcn_ssl_ctxt_t *)data;
     if (c) {
         int i;
-        if (c->crl)
-            X509_STORE_free(c->crl);
         c->crl = NULL;
         if (c->ctx)
             SSL_CTX_free(c->ctx);
@@ -98,12 +98,12 @@ int ssl_callback_ServerNameIndication(SSL *ssl, int *al, tcn_ssl_ctxt_t *c)
     jstring hostname;
     jlong original_ssl_context, new_ssl_context;
     tcn_ssl_ctxt_t *new_c;
-    
+
     // Continue only if the static method exists
     if (sni_java_callback == NULL) {
         return SSL_TLSEXT_ERR_OK;
     }
-    
+
     (*javavm)->AttachCurrentThread(javavm, (void **)&env, NULL);
 
     // Get the host name presented by the client
@@ -112,7 +112,7 @@ int ssl_callback_ServerNameIndication(SSL *ssl, int *al, tcn_ssl_ctxt_t *c)
     // Convert to Java compatible parameters ready for the method call
     hostname = (*env)->NewStringUTF(env, servername);
     original_ssl_context = P2J(c);
-    
+
     new_ssl_context = (*env)->CallStaticLongMethod(env,
                                                    ssl_context_class,
                                                    sni_java_callback,
@@ -131,6 +131,112 @@ int ssl_callback_ServerNameIndication(SSL *ssl, int *al, tcn_ssl_ctxt_t *c)
     return SSL_TLSEXT_ERR_OK;
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L && !defined(LIBRESSL_VERSION_NUMBER)
+/*
+ * This callback function is called when the ClientHello is received.
+ */
+int ssl_callback_ClientHello(SSL *ssl, int *al, void *arg)
+{
+    JavaVM *javavm = tcn_get_java_vm();
+    JNIEnv *env;
+    char *servername = NULL;
+    const unsigned char *pos;
+    size_t len, remaining;
+    tcn_ssl_ctxt_t *c = (tcn_ssl_ctxt_t *) arg;
+    apr_pool_t *subpool = NULL;
+
+    (*javavm)->AttachCurrentThread(javavm, (void **)&env, NULL);
+    // Continue only if the static method exists
+    if (sni_java_callback == NULL) {
+        return SSL_CLIENT_HELLO_SUCCESS;
+    }
+
+    /* We can't use SSL_get_servername() at this earliest OpenSSL connection
+     * stage, and there is no SSL_client_hello_get0_servername() provided as
+     * of OpenSSL 1.1.1. So the code below, that extracts the SNI from the
+     * ClientHello's TLS extensions, is taken from some test code in OpenSSL,
+     * i.e. client_hello_select_server_ctx() in "test/handshake_helper.c".
+     */
+
+    /*
+     * The server_name extension was given too much extensibility when it
+     * was written, so parsing the normal case is a bit complex.
+     */
+    if (!SSL_client_hello_get0_ext(ssl, TLSEXT_TYPE_server_name, &pos,
+                                   &remaining)
+            || remaining <= 2)
+        goto give_up;
+
+    /* Extract the length of the supplied list of names. */
+    len = (*(pos++) << 8);
+    len += *(pos++);
+    if (len + 2 != remaining)
+        goto give_up;
+    remaining = len;
+
+    /*
+     * The list in practice only has a single element, so we only consider
+     * the first one.
+     */
+    if (remaining <= 3 || *pos++ != TLSEXT_NAMETYPE_host_name)
+        goto give_up;
+    remaining--;
+
+    /* Now we can finally pull out the byte array with the actual hostname. */
+    len = (*(pos++) << 8);
+    len += *(pos++);
+    if (len + 2 != remaining)
+        goto give_up;
+
+    /* Use the SNI to switch to the relevant vhost, should it differ from
+     * c->base_server.
+     */
+    if (apr_pool_create(&subpool, c->pool) != APR_SUCCESS) {
+        goto give_up;
+    }
+    servername = apr_pstrmemdup(subpool, (const char *)pos, len);
+
+give_up:
+    if (servername != NULL) {
+        jstring hostname;
+        jlong original_ssl_context, new_ssl_context;
+        tcn_ssl_ctxt_t *new_c;
+
+        hostname = (*env)->NewStringUTF(env, servername);
+        original_ssl_context = P2J(c);
+        new_ssl_context = (*env)->CallStaticLongMethod(env,
+                                                   ssl_context_class,
+                                                   sni_java_callback,
+                                                   original_ssl_context,
+                                                   hostname);
+        (*env)->DeleteLocalRef(env, hostname);
+        if (new_ssl_context != 0 && new_ssl_context != original_ssl_context) {
+            SSL_CTX *ctx;
+            new_c = J2P(new_ssl_context, tcn_ssl_ctxt_t *);
+            ctx = SSL_set_SSL_CTX(ssl, new_c->ctx);
+
+            /* Copied from httpd (modules/ssl/ssl_engine_kernel.c) */
+            SSL_set_options(ssl, SSL_CTX_get_options(ctx));
+            SSL_set_min_proto_version(ssl, SSL_CTX_get_min_proto_version(ctx));
+            SSL_set_max_proto_version(ssl, SSL_CTX_get_max_proto_version(ctx));
+            if ((SSL_get_verify_mode(ssl) == SSL_VERIFY_NONE) ||
+                (SSL_num_renegotiations(ssl) == 0)) {
+                SSL_set_verify(ssl, SSL_CTX_get_verify_mode(ctx), SSL_CTX_get_verify_callback(ctx));
+            }
+            if (SSL_num_renegotiations(ssl) == 0) {
+                SSL_set_session_id_context(ssl,  &(c->context_id[0]), sizeof c->context_id);
+            }
+        }
+    }
+
+    if (subpool != NULL) {
+        apr_pool_destroy(subpool);
+    }
+
+    return SSL_CLIENT_HELLO_SUCCESS;
+}
+#endif /* OPENSSL_VERSION_NUMBER < 0x10101000L */
+
 /* Initialize server context */
 TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
                                             jint protocol, jint mode)
@@ -139,6 +245,7 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
     tcn_ssl_ctxt_t *c = NULL;
     SSL_CTX *ctx = NULL;
     jclass clazz;
+    jclass sClazz;
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L
     jint prot;
 #endif
@@ -150,7 +257,16 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
     }
 
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
-    if (protocol == SSL_PROTOCOL_TLSV1_2) {
+    if (protocol == SSL_PROTOCOL_TLSV1_3) {
+#ifdef HAVE_TLSV1_3
+        if (mode == SSL_MODE_CLIENT)
+            ctx = SSL_CTX_new(TLSv1_3_client_method());
+        else if (mode == SSL_MODE_SERVER)
+            ctx = SSL_CTX_new(TLSv1_3_server_method());
+        else
+            ctx = SSL_CTX_new(TLSv1_3_method());
+#endif
+    } else if (protocol == SSL_PROTOCOL_TLSV1_2) {
 #ifdef HAVE_TLSV1_2
         if (mode == SSL_MODE_CLIENT)
             ctx = SSL_CTX_new(TLSv1_2_client_method());
@@ -184,6 +300,10 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
             ctx = SSL_CTX_new(SSLv3_method());
     } else if (protocol == SSL_PROTOCOL_SSLV2) {
         /* requested but not supported */
+#ifndef HAVE_TLSV1_3
+    } else if (protocol & SSL_PROTOCOL_TLSV1_3) {
+        /* requested but not supported */
+#endif
 #ifndef HAVE_TLSV1_2
     } else if (protocol & SSL_PROTOCOL_TLSV1_2) {
         /* requested but not supported */
@@ -215,6 +335,10 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
         goto init_failed;
     }
 
+#ifdef HAVE_KEYLOG_CALLBACK
+    SSL_callback_add_keylog(ctx);
+#endif
+
     c->protocol = protocol;
     c->mode     = mode;
     c->ctx      = ctx;
@@ -224,7 +348,7 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
         BIO_set_fp(c->bio_os, stderr, BIO_NOCLOSE | BIO_FP_TEXT);
     SSL_CTX_set_options(c->ctx, SSL_OP_ALL);
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
     /* always disable SSLv2, as per RFC 6176 */
     SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
     if (!(protocol & SSL_PROTOCOL_SSLV3))
@@ -239,9 +363,19 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
     if (!(protocol & SSL_PROTOCOL_TLSV1_2))
         SSL_CTX_set_options(c->ctx, SSL_OP_NO_TLSv1_2);
 #endif
+#ifdef HAVE_TLSV1_3
+    if (!(protocol & SSL_PROTOCOL_TLSV1_3))
+        SSL_CTX_set_options(c->ctx, SSL_OP_NO_TLSv1_3);
+#endif
 
-#else /* if OPENSSL_VERSION_NUMBER < 0x10100000L */
+#else /* if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER) */
     /* We first determine the maximum protocol version we should provide */
+#ifdef HAVE_TLSV1_3
+    if (protocol & SSL_PROTOCOL_TLSV1_3) {
+        prot = TLS1_3_VERSION;
+    } else
+/* NOTE the dangling else above: take care to preserve it */
+#endif
     if (protocol & SSL_PROTOCOL_TLSV1_2) {
         prot = TLS1_2_VERSION;
     } else if (protocol & SSL_PROTOCOL_TLSV1_1) {
@@ -259,6 +393,11 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
 
     /* Next we scan for the minimal protocol version we should provide,
      * but we do not allow holes between max and min */
+#ifdef HAVE_TLSV1_3
+    if (prot == TLS1_3_VERSION && protocol & SSL_PROTOCOL_TLSV1_2) {
+        prot = TLS1_2_VERSION;
+    }
+#endif
     if (prot == TLS1_2_VERSION && protocol & SSL_PROTOCOL_TLSV1_1) {
         prot = TLS1_1_VERSION;
     }
@@ -336,6 +475,16 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
     SSL_CTX_set_tlsext_servername_callback(c->ctx, ssl_callback_ServerNameIndication);
     SSL_CTX_set_tlsext_servername_arg(c->ctx, c);
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L && !defined(LIBRESSL_VERSION_NUMBER)
+    /*
+     * The ClientHello callback also allows to retrieve the SNI, but since it
+     * runs at the earliest possible connection stage we can even set the TLS
+     * protocol version(s) according to the selected (name-based-)vhost, which
+     * is not possible at the SNI callback stage (due to OpenSSL internals).
+     */
+    SSL_CTX_set_client_hello_cb(c->ctx, ssl_callback_ClientHello, c);
+#endif
+
     /*
      * Let us cleanup the ssl context when the pool is destroyed
      */
@@ -344,8 +493,12 @@ TCN_IMPLEMENT_CALL(jlong, SSLContext, make)(TCN_STDARGS, jlong pool,
                               apr_pool_cleanup_null);
 
     /* Cache the byte[].class for performance reasons */
-    clazz = (*e)->FindClass(e, "[B");
-    byteArrayClass = (jclass) (*e)->NewGlobalRef(e, clazz);
+    if (stringClass == NULL) {
+        clazz = (*e)->FindClass(e, "[B");
+        byteArrayClass = (jclass) (*e)->NewGlobalRef(e, clazz);
+        sClazz = (*e)->FindClass(e, "java/lang/String");
+        stringClass = (jclass) (*e)->NewGlobalRef(e, sClazz);
+    }
 
     return P2J(c);
 init_failed:
@@ -489,6 +642,65 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCipherSuite)(TCN_STDARGS, jlong ctx,
     return rv;
 }
 
+TCN_IMPLEMENT_CALL(jobjectArray, SSLContext, getCiphers)(TCN_STDARGS, jlong ctx)
+{
+    tcn_ssl_ctxt_t *c = J2P(ctx, tcn_ssl_ctxt_t *);
+    STACK_OF(SSL_CIPHER) *sk;
+    int len;
+    jobjectArray array;
+    SSL_CIPHER *cipher;
+    const char *name;
+    int i;
+    jstring c_name;
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+    SSL *ssl;
+#endif
+
+    UNREFERENCED_STDARGS;
+
+    if (c->ctx == NULL) {
+        tcn_ThrowException(e, "ssl context is null");
+        return NULL;
+    }
+
+    /* Before OpenSSL 1.1.0, get_ciphers() was only available
+     * on an SSL, not for an SSL_CTX. */
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+    ssl = SSL_new(c->ctx);
+    if (ssl == NULL) {
+        tcn_ThrowException(e, "could not create temporary ssl from ssl context");
+        return NULL;
+    }
+
+    sk = SSL_get_ciphers(ssl);
+#else
+    sk = SSL_CTX_get_ciphers(c->ctx);
+#endif
+    len = sk_SSL_CIPHER_num(sk);
+
+    if (len <= 0) {
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+        SSL_free(ssl);
+#endif
+        return NULL;
+    }
+
+    array = (*e)->NewObjectArray(e, len, stringClass, NULL);
+
+    for (i = 0; i < len; i++) {
+        cipher = (SSL_CIPHER*) sk_SSL_CIPHER_value(sk, i);
+        name = SSL_CIPHER_get_name(cipher);
+
+        c_name = (*e)->NewStringUTF(e, name);
+        (*e)->SetObjectArrayElement(e, array, i, c_name);
+    }
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+    SSL_free(ssl);
+#endif
+    return array;
+}
+
+
 TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCARevocation)(TCN_STDARGS, jlong ctx,
                                                           jstring file,
                                                           jstring path)
@@ -502,35 +714,45 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCARevocation)(TCN_STDARGS, jlong ctx
 
     UNREFERENCED(o);
     TCN_ASSERT(ctx != 0);
-    if (J2S(file) == NULL && J2S(path) == NULL)
+    if (J2S(file) == NULL && J2S(path) == NULL) {
         return JNI_FALSE;
+    }
 
     if (!c->crl) {
-        if ((c->crl = X509_STORE_new()) == NULL)
+        if ((c->crl = SSL_CTX_get_cert_store(c->ctx)) == NULL)
             goto cleanup;
     }
     if (J2S(file)) {
         lookup = X509_STORE_add_lookup(c->crl, X509_LOOKUP_file());
         if (lookup == NULL) {
             ERR_error_string(SSL_ERR_get(), err);
-            X509_STORE_free(c->crl);
             c->crl = NULL;
             tcn_Throw(e, "Lookup failed for file %s (%s)", J2S(file), err);
             goto cleanup;
         }
-        X509_LOOKUP_load_file(lookup, J2S(file), X509_FILETYPE_PEM);
+        if (!X509_LOOKUP_load_file(lookup, J2S(file), X509_FILETYPE_PEM)) {
+            ERR_error_string(SSL_ERR_get(), err);
+            c->crl = NULL;
+            tcn_Throw(e, "Load failed for file %s (%s)", J2S(file), err);
+            goto cleanup;
+        }
     }
     if (J2S(path)) {
         lookup = X509_STORE_add_lookup(c->crl, X509_LOOKUP_hash_dir());
         if (lookup == NULL) {
             ERR_error_string(SSL_ERR_get(), err);
-            X509_STORE_free(c->crl);
             c->crl = NULL;
             tcn_Throw(e, "Lookup failed for path %s (%s)", J2S(file), err);
             goto cleanup;
         }
-        X509_LOOKUP_add_dir(lookup, J2S(path), X509_FILETYPE_PEM);
+        if (!X509_LOOKUP_add_dir(lookup, J2S(path), X509_FILETYPE_PEM)) {
+            ERR_error_string(SSL_ERR_get(), err);
+            c->crl = NULL;
+            tcn_Throw(e, "Load failed for path %s (%s)", J2S(file), err);
+            goto cleanup;
+        }
     }
+    X509_STORE_set_flags(c->crl, X509_V_FLAG_CRL_CHECK | X509_V_FLAG_CRL_CHECK_ALL);
     rv = JNI_TRUE;
 cleanup:
     TCN_FREE_CSTRING(file);
@@ -589,7 +811,7 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCACertificate)(TCN_STDARGS,
         c->ca_certs++;
         ca_certs = SSL_CTX_get_client_CA_list(c->ctx);
         if (ca_certs == NULL) {
-            SSL_load_client_CA_file(J2S(file));
+            ca_certs = SSL_load_client_CA_file(J2S(file));
             if (ca_certs != NULL)
                 SSL_CTX_set_client_CA_list(c->ctx, ca_certs);
         }
@@ -638,7 +860,7 @@ TCN_IMPLEMENT_CALL(void, SSLContext, setTmpDH)(TCN_STDARGS, jlong ctx,
         tcn_Throw(e, "Error while configuring DH: no dh param file given");
         return;
     }
-    
+
     bio = BIO_new_file(J2S(file), "r");
     if (!bio) {
         char err[256];
@@ -666,7 +888,7 @@ TCN_IMPLEMENT_CALL(void, SSLContext, setTmpDH)(TCN_STDARGS, jlong ctx,
         TCN_FREE_CSTRING(file);
         return;
     }
-    
+
     DH_free(dh);
     TCN_FREE_CSTRING(file);
 }
@@ -682,7 +904,7 @@ TCN_IMPLEMENT_CALL(void, SSLContext, setTmpECDHByCurveName)(TCN_STDARGS, jlong c
     UNREFERENCED(o);
     TCN_ASSERT(ctx != 0);
     TCN_ASSERT(curveName);
-    
+
     /* First try to get curve by name */
     i = OBJ_sn2nid(J2S(curveName));
     if (!i) {
@@ -690,14 +912,14 @@ TCN_IMPLEMENT_CALL(void, SSLContext, setTmpECDHByCurveName)(TCN_STDARGS, jlong c
         TCN_FREE_CSTRING(curveName);
         return;
     }
-    
+
     ecdh = EC_KEY_new_by_curve_name(i);
     if (!ecdh) {
         tcn_Throw(e, "Can't configure elliptic curve: unknown curve name %s", J2S(curveName));
         TCN_FREE_CSTRING(curveName);
         return;
     }
-    
+
     /* Setting found curve to context */
     if (1 != SSL_CTX_set_tmp_ecdh(c->ctx, ecdh)) {
         char err[256];
@@ -710,8 +932,8 @@ TCN_IMPLEMENT_CALL(void, SSLContext, setTmpECDHByCurveName)(TCN_STDARGS, jlong c
     EC_KEY_free(ecdh);
     TCN_FREE_CSTRING(curveName);
 #else
-	tcn_Throw(e, "Cant't configure elliptic curve: unsupported by this OpenSSL version");
-	return;
+    tcn_Throw(e, "Cant't configure elliptic curve: unsupported by this OpenSSL version");
+    return;
 #endif
 }
 
@@ -890,7 +1112,7 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificate)(TCN_STDARGS, jlong ctx,
     const char *p;
     char err[256];
 #ifdef HAVE_ECC
-    EC_GROUP *ecparams;
+    EC_GROUP *ecparams = NULL;
     int nid;
     EC_KEY *eckey = NULL;
 #endif
@@ -907,7 +1129,7 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificate)(TCN_STDARGS, jlong ctx,
     if (J2S(password)) {
         if (!c->cb_data)
             c->cb_data = &tcn_password_callback;
-        strncpy(c->cb_data->password, J2S(password), SSL_MAX_PASSWORD_LEN);
+        strncpy(c->cb_data->password, J2S(password), SSL_MAX_PASSWORD_LEN - 1);
         c->cb_data->password[SSL_MAX_PASSWORD_LEN-1] = '\0';
     }
     key_file  = J2S(key);
@@ -929,7 +1151,13 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificate)(TCN_STDARGS, jlong ctx,
         }
     }
     else {
-        if ((c->keys[idx] = load_pem_key(c, key_file)) == NULL) {
+        if ((c->keys[idx] = load_pem_key(c, key_file)) == NULL
+#ifndef OPENSSL_NO_ENGINE
+                && (tcn_ssl_engine == NULL ||
+                (c->keys[idx] = ENGINE_load_private_key(tcn_ssl_engine, key_file,
+                                                        NULL, NULL)) == NULL)
+#endif
+                ) {
             ERR_error_string(SSL_ERR_get(), err);
             tcn_Throw(e, "Unable to load certificate key %s (%s)",
                       key_file, err);
@@ -971,6 +1199,7 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificate)(TCN_STDARGS, jlong ctx,
      * If only for PEM files move above to the PEM handling */
     if ((idx == 0) && (dhparams = SSL_dh_GetParamFromFile(cert_file))) {
         SSL_CTX_set_tmp_dh(c->ctx, dhparams);
+        DH_free(dhparams);
     }
 
 #ifdef HAVE_ECC
@@ -985,8 +1214,11 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificate)(TCN_STDARGS, jlong ctx,
         SSL_CTX_set_tmp_ecdh(c->ctx, eckey);
     }
     /*
-     * ...otherwise, configure NIST P-256 (required to enable ECDHE)
+     * ...otherwise, enable auto curve selection (OpenSSL 1.0.2)
+     * or configure NIST P-256 (required to enable ECDHE for earlier versions)
+     * ECDH is always enabled in 1.1.0 unless excluded from SSLCipherList
      */
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
     else {
 #if defined(SSL_CTX_set_ecdh_auto)
         SSL_CTX_set_ecdh_auto(c->ctx, 1);
@@ -995,7 +1227,10 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificate)(TCN_STDARGS, jlong ctx,
         SSL_CTX_set_tmp_ecdh(c->ctx, eckey);
 #endif
     }
+#endif
+    /* OpenSSL assures us that _free() is NULL-safe */
     EC_KEY_free(eckey);
+    EC_GROUP_free(ecparams);
 #endif
     SSL_CTX_set_tmp_dh_callback(c->ctx, SSL_callback_tmp_DH);
 
@@ -1105,6 +1340,7 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificateRaw)(TCN_STDARGS, jlong c
     /*
      * TODO try to read the ECDH curve name from somewhere...
      */
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
 #if defined(SSL_CTX_set_ecdh_auto)
     SSL_CTX_set_ecdh_auto(c->ctx, 1);
 #else
@@ -1112,6 +1348,7 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, setCertificateRaw)(TCN_STDARGS, jlong c
     SSL_CTX_set_tmp_ecdh(c->ctx, eckey);
     EC_KEY_free(eckey);
 #endif
+#endif
 #endif
     SSL_CTX_set_tmp_dh_callback(c->ctx, SSL_callback_tmp_DH);
 cleanup:
@@ -1148,9 +1385,14 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, addChainCertificateRaw)(TCN_STDARGS, jl
         ERR_error_string(SSL_ERR_get(), err);
         tcn_Throw(e, "Error reading certificate (%s)", err);
         rv = JNI_FALSE;
+#if defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x20901000L
+    } else {
+        tcn_Throw(e, "Unable to use Java keystores with LibreSSL");
+#else
     } else if (SSL_CTX_add0_chain_cert(c->ctx, certs) <= 0) {
         ERR_error_string(SSL_ERR_get(), err);
         tcn_Throw(e, "Error adding certificate to chain (%s)", err);
+#endif
         rv = JNI_FALSE;
     }
 
@@ -1158,6 +1400,44 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, addChainCertificateRaw)(TCN_STDARGS, jl
     return rv;
 }
 
+TCN_IMPLEMENT_CALL(jboolean, SSLContext, addClientCACertificateRaw)(TCN_STDARGS, jlong ctx,
+                                                                    jbyteArray javaCert)
+{
+    jsize lengthOfCert;
+    unsigned char *charCert;
+    X509 *cert;
+    const unsigned char *tmp;
+
+    tcn_ssl_ctxt_t *c = J2P(ctx, tcn_ssl_ctxt_t *);
+    jboolean rv = JNI_TRUE;
+    char err[256];
+
+    /* we get the cert contents into a byte array */
+    jbyte* bufferPtr = (*e)->GetByteArrayElements(e, javaCert, NULL);
+    lengthOfCert = (*e)->GetArrayLength(e, javaCert);
+    charCert = malloc(lengthOfCert);
+    memcpy(charCert, bufferPtr, lengthOfCert);
+    (*e)->ReleaseByteArrayElements(e, javaCert, bufferPtr, 0);
+
+    UNREFERENCED(o);
+    TCN_ASSERT(ctx != 0);
+
+    tmp = (const unsigned char *)charCert;
+    cert = d2i_X509(NULL, &tmp, lengthOfCert);
+    if (cert == NULL) {
+        ERR_error_string(SSL_ERR_get(), err);
+        tcn_Throw(e, "Error encoding allowed peer CA certificate (%s)", err);
+        rv = JNI_FALSE;
+    } else if (SSL_CTX_add_client_CA(c->ctx, cert) <= 0) {
+        ERR_error_string(SSL_ERR_get(), err);
+        tcn_Throw(e, "Error adding allowed peer CA certificate (%s)", err);
+        rv = JNI_FALSE;
+    }
+
+    free(charCert);
+    return rv;
+}
+
 static int ssl_array_index(apr_array_header_t *array,
                            const char *s)
 {
@@ -1577,7 +1857,7 @@ TCN_IMPLEMENT_CALL(void, SSLContext, setSessionTicketKeys)(TCN_STDARGS, jlong ct
 }
 
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
 
 /*
  * Adapted from OpenSSL:
@@ -1677,7 +1957,7 @@ static const char* SSL_CIPHER_authentication_method(const SSL_CIPHER* cipher){
     if (cipher == NULL) {
         return "UNKNOWN";
     }
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
     kx = cipher->algorithm_mkey;
     auth = cipher->algorithm_auth;
 #else
@@ -2034,6 +2314,15 @@ TCN_IMPLEMENT_CALL(jboolean, SSLContext, addChainCertificateRaw)(TCN_STDARGS, jl
     return JNI_FALSE;
 }
 
+TCN_IMPLEMENT_CALL(jboolean, SSLContext, addClientCACertificateRaw)(TCN_STDARGS, jlong ctx,
+                                                                    jbyteArray javaCert)
+{
+    UNREFERENCED_STDARGS;
+    UNREFERENCED(ctx);
+    UNREFERENCED(javaCert);
+    return JNI_FALSE;
+}
+
 TCN_IMPLEMENT_CALL(jint, SSLContext, setALPN)(TCN_STDARGS, jlong ctx,
                                               jbyteArray buf, jint len)
 {
