@@ -150,10 +150,11 @@ public final class SessionListener implements ServletContextListener,
      */
     private void log(String message) {
 
-        if (context != null)
+        if (context != null) {
             context.log("SessionListener: " + message);
-        else
+        } else {
             System.out.println("SessionListener: " + message);
+        }
 
     }
 
